/* **********************************************************
 * Copyright (C) 1998-2000 VMware, Inc.
 * All Rights Reserved
 * $Id: modulecall.h,v 1.7.4.3 2002/05/20 21:12:10 hpreg Exp $
 * **********************************************************/


/*
 * modulecall.h
 *
 *        Monitor <-->  Module (kernel driver) interface
 */

#ifndef _MODULECALL_H
#define _MODULECALL_H

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_MONITOR
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#include "includeCheck.h"

#include "x86.h"
#include "iocontrols.h"

/*
 *----------------------------------------------------------------------
 *
 * ModuleCallType --
 *
 *      Enumaration of support calls done by the module
 *
 *----------------------------------------------------------------------
 */

typedef enum ModuleCallType {
   MODULECALL_NONE = 100,            //
   
   /*
    * All I/O interrupts go under MODULECALL_INTR, special-cased
    * in task switching code
    */
   MODULECALL_INTR,                  
   MODULECALL_LOCKPAGE,                  
   MODULECALL_UNLOCKPAGE,                  

   MODULECALL_ISMPNLOCKED, 
   /*
    * calls to user mode
    */
   MODULECALL_USERRETURN,            // return code for USER RPC
   MODULECALL_LAST                   // Number of entries. Must be the last one
} ModuleCallType;

#define MODULECALL_USER_START 300
#define MODULECALL_USER_END   399

#define MODULECALL_CROSS_PAGE_LEN    1
#define MODULECALL_CROSS_PAGE_START_V2		2
#define MODULECALL_CROSS_PAGE_START_GSX1	2
#define MODULECALL_CROSS_PAGE_START_V3		4
#define MODULECALL_CROSS_PAGE_START_V32		4
#define MODULECALL_CROSS_PAGE_START_GSX2	6
#define MODULECALL_CROSS_PAGE_START_GSX25	6

// we don't want it in the final version
//#define USE_BLUE_SCREEN

/*
 *----------------------------------------------------------------------
 *
 * VMCrossPage --
 *
 *      data structure shared between the monitor and the module
 *      that is used for crossing between the two.
 *      Accessible as vm->cross (kernel module) and CROSS_PAGE
 *      (monitor)
 *
 *      Exactly one page long
 *
 *----------------------------------------------------------------------
 */

/*
 * We arrived at 1008 as follows. The first 1024 bytes of the cross
 * page contains the version number (4 bytes) and then the task switch
 * code aligned at a 16 byte boundary. This alignment allows the
 * decoders to get a good start. We allocate 1024 bytes for all this
 * although the task switch code is only about 192 bytes large so
 * that the context information (data) that follows this code fragment
 * is on another 1024 byte chunk of memory. This separation makes the
 * trace cache of the P4 happy. If it any closer, the trace cache
 * gets flushed each time we write to the context info data area. 
 */
#define CONTEXT_SWITCH_CODE_SIZE_GSX25  1008

#define MAX_PASSTHROUGH_PORT 1024
#define NUM_PASSTHROUGH_WORDS (MAX_PASSTHROUGH_PORT / 32)
#define INTERRUPT_REDIRECTION_BITMAP_SIZE 32

#ifdef __GNUC__
 struct ContextInfo {
#else
 typedef  struct ContextInfo {
#endif
    DTRWords    gdtr;
    DTRWords    ldtr;
    DTRWords    idtr;
    uint16 tr;
    uint8 jump[6];
    Task task;   
#ifdef __GNUC__
 } __attribute__ ((packed));

typedef struct ContextInfo ContextInfo;
#else
 } ContextInfo;
#endif

#ifdef __GNUC__
 struct VMCrossPageGSX25 {
#else
 typedef struct VMCrossPageGSX25 {
#endif
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_GSX25];

   /*
    * iobitMap must immediately follow the interrupt redirection
    * bitmap, which must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   struct VMCrossPageGSX25 *hostVA;

   PTE    hostSharedPTE;

   uint32 hostContextVA;


   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;
  
   MPN crossMonPageTableMPN; 

   /* PDE that points to switch page table */
   PDE  switchPDE;
   PAE_PDE paeSwitchPDE; 

   LA monitorLinearAddr;      // base location of monitor in linear space
  
   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA crosspageLinearAddr;

   //
   // ModuleCall parameters
   //
   ModuleCallType moduleCallType;
   uintptr_t args[4];
   uintptr_t retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

   Bool monUsesSysenter;

#if defined(WINNT_DDK) && defined(VMX86_DEBUG)
   void *vmPtr;  /* Monitor to VmDriver structure, useful for debugging */
#endif

#ifdef __GNUC__
 } __attribute__ ((packed));

typedef struct VMCrossPageGSX25 VMCrossPageGSX25;
#else
 } VMCrossPageGSX25;
#endif

#define monUsesSysenterGSX25(x) ((x)->monUsesSysenter)

/*
 * We hardcode these constants here instead of using the sizeof() and
 * offsetof() to catch compiler inconsistencies. This file is compiled
 * once onsite to build the monitor, and then once at the customer site
 * to build the drivers.
 *
 * An alternate scheme, and preferred scheme, would be have the makefile
 * generate them, thereby relieving the developer of the error prone
 * chore of updating them each time VMCrossPage is modified. 
 *
 * Pratap 12/01.
 */
#define MODULECALL_CROSSPAGE_SIZE_GSX25     0x5fd
#define MODULECALL_CROSSPAGE_LAOFFSET_GSX25 0x5f4

#define CROSSPAGE_VERSION_GSX25 (0x1777 + VERSION_MAGIC)

#define CONTEXT_SWITCH_CODE_SIZE_GSX2  1008

#define MAX_PASSTHROUGH_PORT 1024
#define NUM_PASSTHROUGH_WORDS (MAX_PASSTHROUGH_PORT / 32)

#ifdef __GNUC__
 struct VMCrossPageGSX2 {
#else
 typedef struct VMCrossPageGSX2 {
#endif
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_GSX2];

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   struct VMCrossPageGSX2 *hostVA;

   PTE    hostSharedPTE;

   uint32 hostContextVA;


   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;
  
   MPN crossMonPageTableMPN; 

   /* PDE that points to switch page table */
   PDE  switchPDE;
   PAE_PDE paeSwitchPDE; 

   LA monitorLinearAddr;      // base location of monitor in linear space
  
   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA crosspageLinearAddr;

   //
   // ModuleCall parameters
   //
   ModuleCallType moduleCallType;
   uintptr_t args[4];
   uintptr_t retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

   Bool monUsesSysenter;

#if defined(WINNT_DDK) && defined(VMX86_DEBUG)
   void *vmPtr;  /* Monitor to VmDriver structure, useful for debugging */
#endif

#ifdef __GNUC__
 } __attribute__ ((packed));

typedef struct VMCrossPageGSX2 VMCrossPageGSX2;
#else
 } VMCrossPageGSX2;
#endif

#define monUsesSysenterGSX2(x) ((x)->monUsesSysenter)

/*
 * We hardcode these constants here instead of using the sizeof() and
 * offsetof() to catch compiler inconsistencies. This file is compiled
 * once onsite to build the monitor, and then once at the customer site
 * to build the drivers.
 *
 * An alternate scheme, and preferred scheme, would be have the makefile
 * generate them, thereby relieving the developer of the error prone
 * chore of updating them each time VMCrossPage is modified. 
 *
 * Pratap 12/01.
 */
#define MODULECALL_CROSSPAGE_SIZE_GSX2     0x5dd
#define MODULECALL_CROSSPAGE_LAOFFSET_GSX2 0x5b4

#define CROSSPAGE_VERSION_GSX2 (0x1777 + VERSION_MAGIC)

#define CONTEXT_SWITCH_CODE_SIZE_V3	192

#ifdef __GNUC__
 struct VMCrossPageV3 {
#else
 typedef struct VMCrossPageV3 {
#endif
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_V3];

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   struct VMCrossPageV3 * hostVA;
   PTE    hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   PDE  switchPDE;            // PDE that points to switch page table


   LA monitorLinearAddr;      // base location of monitor in linear space
   LA crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallType moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#if defined(WINNT_DDK) && defined(VMX86_DEBUG)
   void *vmPtr;  /* Monitor to VmDriver structure, useful for debugging */
#endif

#ifdef __GNUC__
 } __attribute__ ((packed));

typedef struct VMCrossPageV3 VMCrossPageV3;
#else
 } VMCrossPageV3;
#endif

#define monUsesSysenterV3(x) (0)

#define MODULECALL_CROSSPAGE_SIZE_V3     0x294
#define MODULECALL_CROSSPAGE_LAOFFSET_V3 0x26c

#define CONTEXT_SWITCH_CODE_SIZE_V32	1008

#ifdef __GNUC__
 struct VMCrossPageV32 {
#else
 typedef struct VMCrossPageV32 {
#endif
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
   uint32 _pad[3];

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_V32];

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   struct VMCrossPageV32 * hostVA;
   PTE    hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   PDE  switchPDE;            // PDE that points to switch page table


   LA monitorLinearAddr;      // base location of monitor in linear space
   LA crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallType moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#if defined(WINNT_DDK) && defined(VMX86_DEBUG)
   void *vmPtr;  /* Monitor to VmDriver structure, useful for debugging */
#endif

#ifdef __GNUC__
 } __attribute__ ((packed));

typedef struct VMCrossPageV32 VMCrossPageV32;
#else
 } VMCrossPageV32;
#endif

#define monUsesSysenterV32(x) (0)

#define MODULECALL_CROSSPAGE_SIZE_V32     0x5d0
#define MODULECALL_CROSSPAGE_LAOFFSET_V32 0x5a8

#define CROSSPAGE_VERSION_V32 (0x1776 + VERSION_MAGIC)

#ifndef VMX86_SERVER
#define MONITOR_LINEAR_START            (CROSS_PAGE->monitorLinearAddr)
#ifdef MONITOR
#define CROSS_PAGE  ((VMCrossPage *)(MONITOR_BASE_VADDR + VPN_2_VA(CROSS_PAGE_START)))
#endif
#else
#ifdef VMX86_8MB_MONITOR
#define MONITOR_LINEAR_START		0xff800000
#else
#define MONITOR_LINEAR_START		0xffc00000
#endif
#endif

#define NULLPAGE_LINEAR_START           (MONITOR_LINEAR_START + PAGE_SIZE*NULLPAGE_START)
   

struct passthrough_iorange {
  unsigned short ioBase;	/* Base of range to pass through */
  unsigned short numPorts;	/* Length of range */
};

#endif
